/* 
 * The contents of this file are subject to the GrammarSoft Public
 * License Version 1.0 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.grammarsoft.com/GSPL
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is the VISL CG Parser.
 * 
 * The Initial Developer of the Original Code is Martin Carlsen, 
 * GrammarSoft, and SDU Odense University.  
 * Portions created by Martin Carlsen, Grammarsoft, and SDU Odense 
 * University are Copyright (C) 1999-2000 Martin Carlsen, GrammarSoft, 
 * and SDU Odense University.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * Alternatively, the contents of this file may be used under the
 * terms of the GNU General Public License Version 2 or later (the
 * "GPL"), in which case the provisions of the GPL are applicable 
 * instead of those above.  If you wish to allow use of your 
 * version of this file only under the terms of the GPL and not to
 * allow others to use your version of this file under the GSPL,
 * indicate your decision by deleting the provisions above and
 * replace them with the notice and other provisions required by
 * the GPL.  If you do not delete the provisions above, a recipient
 * may use your version of this file under either the GSPL or the
 * GPL.
 */
#include <vector>
#include <list>
#include <string>

#include "cg.data_structures.hh"
#include "CGCompiler.hh"
#include "Debug.hh"

TagRegister * CGCompiler::tag_register = TagRegister::instance();
Debug * CGCompiler::debug = Debug::instance();

Combined * CGCompiler::correct = 0;
Combined * CGCompiler::correctTag()
{
  if (correct == 0) {
    correct = new Combined();
    string * ctag = new string("<Correct!>");
    correct->push_back(TagRegister::instance()->add(*ctag));  }
  return correct;
}

/**
 * Reorder rules, according to the preferred-targets list.
 * BUG: The measures taken here are unlikely to be sufficient to ensure 
 *      the same priorities as CG-2.
 */
void CGCompiler::reorderRules()
{
  list<list<Constraint*>*>* reorderedsections = new list<list<Constraint*>*>();
  
  // for each section
  for(list<list<Constraint*>*>::iterator 
	ci = constraintsections->begin(),
	cz = constraintsections->end();
      ci!=cz; ++ci)
    {
      list<Constraint*>* constraints = *ci;
      int nSelectOrdered = 0;
      int nSelectUnordered = 0;
      int nRemoveOrdered = 0;
      int nRemoveUnordered = 0;
      
      hash_map<int,list<Constraint*>* >* select_ordered = new hash_map<int,list<Constraint*>* >();
      hash_map<int,list<Constraint*>* >* remove_ordered = new hash_map<int,list<Constraint*>* >();
      list<Constraint*>* remove_unordered = new list<Constraint*>();
      list<Constraint*>* select_unordered = new list<Constraint*>();
      
      // for each preferred target, initialize the list of 
      for(taglist<int>::iterator
	    pi = preferredlist->begin(),
	    pz = preferredlist->end();
	  pi!=pz; ++pi)
	{
	  (*select_ordered)[*pi] = new list<Constraint*>();
	  (*remove_ordered)[*pi] = new list<Constraint*>();
	}
      
      // cerr << constraints->size() << " rules in section.\n";
      
      // for each rule in section
      for(list<Constraint*>::iterator 
	    i = constraints->begin(),
	    z = constraints->end();
	  i!=z; ++i)
	{
	  Constraint * constraint = *i;
	  bool isOrdered = false;
	  
	  // for each target
	  for(taglist<int>::iterator
		pi = preferredlist->begin(),
		pz = preferredlist->end();
	      pi!=pz; ++pi)
	    {
	      Combined * tag = new Combined();
	      tag->push_back(*pi);
	      
	      if(constraint->getTarget()->targets(tag))
		{
		  if(constraint->getOp()==op_remove)
		    {
		      ((*remove_ordered)[*pi])->push_back(*i);
		      ++nRemoveOrdered;
		    }
		  else
		    {
		      ((*select_ordered)[*pi])->push_back(*i);
		      ++nSelectOrdered;
		    }
		  break;
		}
	      delete tag;
	    }
	  if(!isOrdered) 
	    if(constraint->getOp()==op_remove)
	      {
		remove_unordered->push_back(*i);
		++nRemoveUnordered;
	      }
	    else
	      {
		select_unordered->push_back(*i);
		++nSelectUnordered;
	      }
	}
      
      //cerr << "nRemoveOrdered="<<nRemoveOrdered<<
      //" nSelectOrdered="<<nSelectOrdered<<
      //" nRemoveUnordered="<<nRemoveUnordered<<
      //" nSelectUnordered="<<nSelectUnordered<<"\n";
      
      // clear section and insert rules in the preferred order.
      
      constraints->clear();
      
      // 1. SELECT rules which target preferred targets.
      //    1.1. Most preferred targets first.
      //         1.1.1. Order of appearance.
      for(taglist<int>::iterator
	    pi = preferredlist->begin(),
	    pz = preferredlist->end();
	  pi!=pz; ++pi)
	{
	  constraints->insert(constraints->end(), 
			      ((*select_ordered)[*pi])->begin(), 
			      ((*select_ordered)[*pi])->end());
	  delete ((*select_ordered)[*pi]);
	}
      
      // 2. SELECT rules which target non-preferred targets.
      //    2.1. Order of appearance.
      constraints->insert(constraints->end(), 
			  select_unordered->begin(), 
			  select_unordered->end());
      delete select_unordered;
      
      // 3. REMOVE rules which target non-preferred targets.
      //    3.1. Order of appearance.
      constraints->insert(constraints->end(), 
			  remove_unordered->begin(), 
			  remove_unordered->end());
      delete remove_unordered;
      
      // 4. REMOVE rules which target preferred targets.
      //    4.1. Least preferred targets first.
      //         4.1.1. Order of appearance.
      for(taglist<int>::reverse_iterator
	    pi = preferredlist->rbegin(),
	    pz = preferredlist->rend();
	  pi!=pz; ++pi)
	{
	  constraints->insert(constraints->end(), 
			      ((*remove_ordered)[*pi])->begin(), 
			      ((*remove_ordered)[*pi])->end());
	  delete ((*remove_ordered)[*pi]);
	}
      reorderedsections->push_back(constraints);
      delete remove_ordered;
      delete select_ordered;
    }
  constraintsections->clear();
  constraintsections->insert(constraintsections->end(), 
			     reorderedsections->begin(),
			     reorderedsections->end());
  delete reorderedsections;
}



/**
 * Apply disambiguation rules.
 * 
 * (See 2.4. and 2.6.)
 * allow_unsafe: whether to allow the last reading of a cohort to be removed.
 * sections: the number of constraint sections to apply.
 *           sections < 0 indicates that all sections must be applied.
 */
void CGCompiler::applyConstraints(vector<Cohort*>* cohorts, bool allow_unsafe = false, int sections = -1)
{
  if( sections<0 ) sections = constraintsections->size();
  //cerr << "sections=" << sections << "\n";

  /*
   * (2.6.1):
   * "First, the first section is applied alone, and when it cannot 
   *  disambiguate anymore, the first and second sections are applied
   *  together, etc."
   * (2.6.4):
   * "Basically, the cohorts are disambiguated in an order from left to right.
   *  When all the cohorts have been disambiguated, the second round will be
   *  done and so on, until nothing further can be disambiguated."
   */
  // while (wasDisambiguated)
  //   for each cohort
  //     SECTION: for each section
  //       for each rule in section
  //         for each reading in cohort
  //           rule->disambiguate(reading)
  //           if( !cohort->isDisambiguous() ) break SECTION;
  
  bool wasDisambiguated = true;
  int sectionsToApply = 1;
  
  // continue applying disambiguation rules
  // until no rule was successfully applied to the input 
  // during the previous iteration.
  while(wasDisambiguated || sectionsToApply < sections)
    {
      if(!wasDisambiguated) ++sectionsToApply;
      if(sectionsToApply > sections) break;
      wasDisambiguated = false;
      
      // for each cohort
      for(unsigned int pos = 1; // no sense in trying to disambiguate cohort 0 (>>>)
	  pos < cohorts->size(); 
	  ++pos)
	{
	  Cohort * cohort = (*cohorts)[pos];
	  list<Combined*> * readings = cohort->getReadings();
	  
	  // don't try to disambiguate non-ambiguous cohorts
	  // the --unsafe flag (allow_unsafe) overrides this
	  bool isAmbiguous = 
	    allow_unsafe
	    ? (readings->size() > 0)
	    : (readings->size() > 1);
	  if(!isAmbiguous) continue; 
	  
	  //cerr << "DISAMBIGUATING: "<<(*cohort) << "\n";
	  
	  // for each section, apply rules:
	  int section = 1;
	  for(list<list<Constraint*>*>::iterator 
		ci = constraintsections->begin(),
		cz = constraintsections->end();
	      ci!=cz && isAmbiguous; ++ci)
	    {
	      if(section > sectionsToApply /* || section > sections */ ) break;
	      //cerr << "section=" << section << "\n";
	      ++section;
	      list<Constraint*>* constraints = *ci;
	      
	      // for each rule in section
	      for(list<Constraint*>::iterator 
		    i = constraints->begin(),
		    z = constraints->end();
		  i!=z && isAmbiguous; ++i)
		{
		  Constraint * constraint = *i;
		  
		  //cerr <<"RULE: "<< (*constraint) << "\n";
		  
		  // for each reading in cohort
		  list<Combined*>::iterator
		    ri = readings->begin(),
		    rz = readings->end();
		  while(ri!=rz && isAmbiguous)
		    {
		      list<Combined*>::iterator tmpri = ri;
		      ++ri;
		      
		      // attempt disambiguation.
		      // (tmpri may be erased, that's why we operate on a copy
		      //  and increase ri _before_ disambiguating.)
		      if(!constraint->disambiguate(cohorts, pos, tmpri)) continue;
		      
		      // tracing output
		      if(!debug->minimal) tcout << (*constraint) << "\n";
		      if(debug->verbose) tcout << (*cohort) << "\n";
		      
		      // allow another round of disambiguation attempts.
		      wasDisambiguated = true;
		      
		      // stop further disambiguation attempts (for this cohort)
		      // if the cohort is not ambiguous anymore.
		      isAmbiguous = 
			allow_unsafe
			? (readings->size() > 0)
			: (readings->size() > 1);
		      
		      // SELECT messes up the iterator (ri) 
		      // just break in case of a SELECT operation
		      // (even in unsafe mode).
		      if(constraint->getOp() == op_select) break;
		    }
		}
	    }
	}
    }
}

/**
 * Apply mapping rules.
 *
 * (See 2.5.)
 */
void CGCompiler::applyMapping(vector<Cohort*>* cohorts)
{
  // for each cohort
  for(unsigned int pos = 1; // no sense in trying to map cohort 0 (<<<)
      pos < cohorts->size(); 
      ++pos)
    {
      Cohort * cohort = (*cohorts)[pos];
      
      // for each reading in cohort
      for(list<Combined*>::iterator
	    ri = cohort->getReadings()->begin(),
	    rz = cohort->getReadings()->end();
	  ri!=rz; ++ri)
	{
	  Combined * reading = *ri;
	  
	  // for each mapping rule
	  // (applied in order of appearance in the rules file)
	  for(list<Mapping*>::iterator
		mi = mappings->begin(),
		mz = mappings->end();
	      mi!=mz; ++mi)
	    {
	      Mapping * mapping = *mi;
	      
	      // attempt to apply mapping rule
	      if(! mapping->map(cohorts, pos, reading) ) continue;
	      
	      // tracing output.
	      if(!debug->minimal) tmout << (*mapping) << "\n";
	      if(debug->verbose)  tmout << (*cohort) << "\n";
	      
	      // (2.5.1: "After a MAP operation is applied no other modifications are done.")
	      // (2.5.1: "No other mappings are done to the reading after the REPLACE operation.")
	      MappingOp op = mapping->getOp();
	      if(op == op_map) break;
	      if(op == op_replace) break;
	    }
	}
    }
}

/**
 * Apply correction rules.
 */
void CGCompiler::applyCorrections(vector<Cohort*>* cohorts)
{
  // for each cohort
  for(unsigned int pos = 1; // no sense in trying to correct cohort 0 (<<<)
      pos < cohorts->size(); 
      ++pos)
    {
      Cohort * cohort = (*cohorts)[pos];
      int nReading = 0;
      
      // for each reading in cohort
      for(list<Combined*>::iterator
	    ri = cohort->getReadings()->begin(),
	    rz = cohort->getReadings()->end();
	  ri!=rz; ++ri)
	{
	  ++nReading;
	  
	  // for each correction rule
	  // (applied in order of appearance in the rules file)
	  for(list<Correction*>::iterator
		ci = corrections->begin(),
		cz = corrections->end();
	      ci!=cz; ++ci)
	    {
	      Correction * correction = *ci;
	      
	      // only apply APPENDs to the first reading
	      // (i.e. once per cohort)
	      if(nReading > 1 && correction->getOp()==op_append) continue;
	      
	      //cout << "trying " << (*correction) << "\n";
	      
	      // attempt to apply correction rule
	      if(! correction->correct(cohorts, pos, ri) ) continue;
	      
	      // tracing output.
	      if(!debug->minimal) tmout << (*correction) << "\n";
	      if(debug->verbose)  tmout << (*cohort) << "\n";
	    }
	}
    }
}

/**
 * Turn readings with multiple @-tags into multiple reading with each one @-tag.
 * E.g.: 
 *      A B @C D @E
 * becomes
 *      A B @C D
 *      A B @E
 *
 * (See 2.5.3.)
 * Note: The @-marker may be set to something else than @ by use of 
 * the --prefix command line argument. The TagRegister knows which tags
 * are marked (TagRegister::isSpecial).
 */
void CGCompiler::expandCohorts(vector<Cohort*>* cohorts)
{
  // for each cohort
  for(unsigned int pos = 0; 
      pos < cohorts->size(); 
      ++pos)
    {
      Cohort * cohort = (*cohorts)[pos];
      
      // for each reading in cohort
      for(list<Combined*>::iterator
	    ri = cohort->getReadings()->begin(),
	    rz = cohort->getReadings()->end();
	  ri!=rz; ++ri)
	{
	  Combined * reading = *ri;
	  Combined::iterator firstAtTag = reading->end();
	  
	  // for each tag in the reading
	  for(Combined::iterator
		i=reading->begin(),
		z=reading->end();
	      i!=z; ++i)
	    {
	      if( (tag_register->isSpecial(*i)) )
		{
		  // is this the 2nd @-tag in the current reading ?
		  if(firstAtTag != reading->end())
		    {
		      // new reading to add to the cohort.
		      Combined * newReading = new Combined();
		      // copy tags from the reading to the new one.
		      // all tags until the one before the second @-tag are copied
		      // thus, from (A B @C D @E), (A B @C D) is copied.
		      newReading->insert(newReading->begin(), reading->begin(), i);
		      // add the new reading to the cohort.
		      cohort->getReadings()->push_front(newReading);
		      // delete the copied @-tag 
		      // (and any non-@-tags between that and the second @-tag.
		      // thus, from (A B @C D @E), the tags @C D are deleted.
		      reading->erase(firstAtTag, i);
		    }
		  // the @-tag at this position will be considered the first.
		  firstAtTag = i;
		}
	    }
	}
    }
}
