/* 
 * The contents of this file are subject to the GrammarSoft Public
 * License Version 1.0 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.grammarsoft.com/GSPL
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is the VISL CG Parser.
 * 
 * The Initial Developer of the Original Code is Martin Carlsen, 
 * GrammarSoft, and SDU Odense University.  
 * Portions created by Martin Carlsen, Grammarsoft, and SDU Odense 
 * University are Copyright (C) 1999-2000 Martin Carlsen, GrammarSoft, 
 * and SDU Odense University.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * Alternatively, the contents of this file may be used under the
 * terms of the GNU General Public License Version 2 or later (the
 * "GPL"), in which case the provisions of the GPL are applicable 
 * instead of those above.  If you wish to allow use of your 
 * version of this file only under the terms of the GPL and not to
 * allow others to use your version of this file under the GSPL,
 * indicate your decision by deleting the provisions above and
 * replace them with the notice and other provisions required by
 * the GPL.  If you do not delete the provisions above, a recipient
 * may use your version of this file under either the GSPL or the
 * GPL.
 */
#ifndef _CGCompiler_hh_
#define _CGCompiler_hh_

#include <map>
#include <string>
#include <list>
#include <set>  

#include "cg.data_structures.hh"
#include "Debug.hh"

class CGCompiler
{
private:
  static TagRegister * tag_register;
  static Debug * debug;
  static Combined * correct;
  
  taglist<int> * delimiterlist;
  taglist<int> * preferredlist;
  map<string,Set*> * sets;
  list<Correction*> * corrections;
  list<list<Constraint*>*> * constraintsections;
  list<Mapping*> * mappings;
  
public:
  taglist<int> * getDelimiters() { return delimiterlist; }
  taglist<int> * getPreferredTargets() { return preferredlist; }
  map<string,Set*> * getSets() { return sets; }
  list<list<Constraint*>*> * getConstraints() { return constraintsections; }
  list<Correction*> * getCorrections() { return corrections; }
  list<Mapping*> * getMappings() { return mappings; }
  
  CGCompiler(taglist<int> * dlist, 
	     taglist<int> * plist, 
	     map<string,Set*> * s, 
	     list<Correction*> * c,
	     list<list<Constraint*>*> * csections, 
	     list<Mapping*> * m
	     ) :
    delimiterlist(dlist), 
    preferredlist(plist), 
    sets(s), 
    corrections(c),
    constraintsections(csections), 
    mappings(m)
    {
      tag_register = TagRegister::instance();
    }
  
  void reorderRules();
  void applyConstraints(vector<Cohort*>* cohorts, bool allow_unsafe = false,
			int sections = -1);
  void applyMapping(vector<Cohort*>* cohorts);
  void applyCorrections(vector<Cohort*>* cohorts);
  static void CGCompiler::expandCohorts(vector<Cohort*>* cohorts);
  static Combined * correctTag();
};

#endif
